from ... import base
from ....cms import SQLContent
from ....cms import declared_attr, Column, Integer, String, Boolean, DateTime, ForeignKey, relationship

class User(base.User, SQLContent):
    __tablename__ = "auth_user"
    userid = 'name' # change userid to 'email', for example, specify email as user identifier
    #__table_args__ = {'extend_existing': True}

    @declared_attr
    def id(cls):
        return Column(Integer, primary_key=True)

    @declared_attr
    def name(cls):
        return Column(String(50),unique=True,nullable=False)

    @declared_attr
    def authd(cls):
        return Column(String(160),unique=False,nullable=False)

    @declared_attr
    def is_active(cls):
        return Column(Boolean(),nullable=False) #used to disable accounts



from ..cms.base import ContentManager
from ..cms.blocks import ContentLstBlock, ContentAddBlock, ContentModBlock, ContentDelBlock
from ..user.access_policies import privilege_required
from ..utils import ensure_type
from .. import base, tags, callbacks

class Arch(base.Arch):

    def __init__(self, content_manager, arch_name, **kwargs):
        super().__init__(arch_name, **kwargs)
        ensure_type(content_manager, ContentManager, 'content_manager')

        SELECT = 'select'
        INSERT = 'insert'
        UPDATE = 'update'
        DELETE = 'delete'

        rb = ContentLstBlock(SELECT, content_manager,
                access_policy=privilege_required(f'{arch_name}.{SELECT}'))
        self.add_route_block(rb)

        rb = ContentAddBlock(INSERT, content_manager, reroute_to=SELECT,
                access_policy=privilege_required(f'{arch_name}.{INSERT}'))
        self.add_route_block(rb)

        rb = ContentModBlock(UPDATE, content_manager, reroute_to=SELECT,
                access_policy=privilege_required(f'{arch_name}.{UPDATE}'))
        self.add_route_block(rb)

        rb = ContentDelBlock(DELETE, content_manager, reroute_to=SELECT,
                access_policy=privilege_required(f'{arch_name}.{DELETE}'))
        self.add_route_block(rb)

        for rb in self.route_blocks.values():
            rb.set_custom_callback(tags.SUCCESS, callbacks.default_success)
            rb.set_custom_callback(tags.USER_ERROR, callbacks.default_user_error)
            rb.set_custom_callback(tags.INTEGRITY_ERROR, callbacks.default_int_error)



from .base import User

from ....auth.base import Auth, AuthManager
from ....cms.volatile.procmem import ContentManager

class UserManager(AuthManager, ContentManager):

    def __init__(self, auth_class, user_class=User):
        if not issubclass(auth_class, Auth):
            raise TypeError(f'{auth_class} should be a subclass of {Auth}.')
        if not issubclass(user_class, User):
            raise TypeError(f'{user_class} should be a subclass of {User}.')

        class AuthUser(auth_class, user_class):
            pass

        self.content_class = AuthUser
        self.data = {}

    def select_user(self, userid):
        return self.select_one(userid)



# basic user class to work with flask-login

from .base import User as BaseUser
from .base import Role as BaseRole
from .base import no_role

from .volatile.procmem import User as ProcMemUser
from .volatile.procmem import UserManager as ProcMemUserManager

from .persist.sql import User as SQLUser
from .persist.sql import UserWithRole as SQLUserWithRole
from .persist.sql import Role as SQLRole
from .persist.sql import UserManager as SQLUserManager



from flask import abort
from flask_login import current_user, login_required
from functools import wraps

from .base import Role

def privilege_required(privilege):
    def outer_dec(fn):
        @wraps(fn)
        def decorated_view(*args, **kwargs):
            if not current_user.is_authenticated:
                # not logged in
                abort(401)

            cu_role = current_user.get_role()
            if not isinstance(cu_role, Role):
                # role is not a base role
                abort(403)

            if not cu_role.has_privilege(privilege):
                # does not have the right privilege
                abort(403)

            # is logged in, has a role, and has the required privilege
            return fn(*args, **kwargs)
        return decorated_view
    return outer_dec

def rolename_required(rolename):
    def outer_dec(fn):
        @wraps(fn)
        def decorated_view(*args, **kwargs):
            if not current_user.is_authenticated:
                # not logged in
                abort(401)

            cu_role = current_user.get_role()
            if not isinstance(cu_role, Role):
                # role is not a base role
                abort(403)

            if not cu_role.name == rolename:
                # does not have the right privilege
                abort(403)

            # is logged in, has a role, and has the required privilege
            return fn(*args, **kwargs)
        return decorated_view
    return outer_dec



from .auth import Arch as AuthArch
from .user_manage import Arch as UserManageArch
from .cms import Arch as CMSArch

from .password import Auth as PasswordAuth

from .sql import default_base as default_sql_declarative_base
from .sql import DefaultRole as DefaultSQLRole



# base object for the content management system
# this system also handles user
import datetime

class Content:
    '''ancestor of all content managed by a ContentManager'''

    def get_id(self):
        if hasattr(self, 'id'):
            return self.id
        else:
            return None

    @classmethod
    def create(cls, data):
        raise NotImplementedError(f'create callback on {cls.__name__} not implemented.')

    def update(self, data):
        raise NotImplementedError(f'update callback on {self.__class__.__name__} not implemented.')

    def delete(self, data):
        raise NotImplementedError(f'delete callback on {self.__class__.__name__} not implemented.')

    @classmethod
    def parse_id(cls, data):
        return data['id']


class ContentManager:

    def __init__(self, content_class):
        if not issubclass(content_class, Content):
            raise TypeError(f'{content_class} should be a subclass of {Content}.')
        self.content_class = content_class

    def construct(self, *args, **kwargs):
        return self.content_class(*args, **kwargs)

    def create(self, data, creator=None):
        nc = self.content_class.create(data)
        if isinstance(creator, Content):
            nc.creator_id = creator.get_id()
        nc.created_on = datetime.datetime.now()
        return nc

    def query(self, qargs):
        cid = self.content_class.parse_id(qargs)
        c = self.select_one(cid)
        return c

    def modify(self, qargs, data, modifier=None):
        ec = self.query(qargs)
        ec.update(data)
        if isinstance(modifier, Content):
            ec.modifier_id = modifier.get_id()
        ec.modified_on = datetime.datetime.now()
        return ec

    # get queries
    def select(self, query):
        # specific query
        raise NotImplementedError(f'select method on {self.__class__.__name__} not implemented.')

    def select_all(self):
        # list contents
        raise NotImplementedError(f'select_all method on {self.__class__.__name__} not implemented.')

    def select_one(self, id):
        # select content by id
        raise NotImplementedError(f'select_one method on {self.__class__.__name__} not implemented.')

    # insert/update/delete queries
    def insert(self, nd):
        # insert a new content
        raise NotImplementedError(f'insert method on {self.__class__.__name__} not implemented.')

    def update(self, nd):
        # update a content
        raise NotImplementedError(f'update method on {self.__class__.__name__} not implemented.')

    def delete(self, nd):
        # delete a content
        raise NotImplementedError(f'delete method on {self.__class__.__name__} not implemented.')

    # persistence method
    def commit(self):
        # persist changes and synchronize
        raise NotImplementedError(f'commit method on {self.__class__.__name__} not implemented.')

    def rollback(self):
        # rollback changes (encountered an exception)
        raise NotImplementedError(f'rollback method on {self.__class__.__name__} not implemented.')

    def parse_id(self, data):
        cid = self.content_class.parse_id(data)
        return cid



import copy
from flask import request
from flask_login import login_user, logout_user, current_user, login_required

from .. import tags, exceptions
from .base import Auth, AuthManager
from ..cms import ContentManageBlock, ContentPrepExecBlock, ContentLstBlock, ContentAddBlock, ContentModBlock, ContentDelBlock
from ..utils import ensure_type

class LogoutBlock(ContentManageBlock):

    def __init__(self, keyword, auth_manager, **kwargs):
        super().__init__(keyword, auth_manager, **kwargs)
        ensure_type(auth_manager, AuthManager, 'auth_manager')
        self.auth_manager = self.content_manager

    def view(self):
        if not current_user.is_authenticated:
            # user is not authenticated
            return self.reroute()
        identifier = current_user.get_id()
        logout_user()
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()

class PrepExecBlock(ContentPrepExecBlock):

    def __init__(self, keyword, auth_manager, **kwargs):
        super().__init__(keyword, auth_manager, **kwargs)
        ensure_type(auth_manager, AuthManager, 'auth_manager')
        self.auth_manager = self.content_manager

class LoginBlock(PrepExecBlock):

    def prepare(self):
        identifier, auth_data = self.auth_manager.parse_login(request.form.copy())
        user = self.auth_manager.select_user(identifier)
        if not isinstance(user, Auth):
            raise exceptions.INVALID_CREDS

        if not user.auth(auth_data):
            raise exceptions.INVALID_CREDS

        return (identifier, user)

    def execute(self, identifier, user):
        # auth success
        login_user(user)
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()

class RegisterBlock(PrepExecBlock):

    def prepare(self):
        user = self.auth_manager.register(request.form.copy())
        return (user,)

    def execute(self, user):
        # insert new user
        identifier = self.auth_manager.insert(user)
        self.auth_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()

class RenewBlock(PrepExecBlock):

    @property
    def default_access_policy(self):
        return login_required

    def prepare(self):
        # shallow copy a user (as opposed to deepcopy)
        user = copy.deepcopy(current_user)
        identifier = user.get_id()
        # update current user from request
        user.renew(request.form.copy())
        logout_user() # logout user from flask-login
        return (identifier, user)

    def execute(self, identifier, user):
        # insert the updated new user
        login_user(user) # login the copy
        self.auth_manager.update(user)
        self.auth_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()

class ResetBlock(PrepExecBlock):

    def prepare(self):
        identifier = self.auth_manager.parse_reset(request.form.copy())
        user = self.auth_manager.select_user(identifier)
        if not isinstance(user, Auth):
            raise exceptions.INVALID_CREDS
        user.reset(request.form.copy())  # reset auth data
        return (identifier, user)

    def execute(self, identifier, user):
        self.auth_manager.update(user)
        self.auth_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()

class RemoveBlock(PrepExecBlock):

    @property
    def default_access_policy(self):
        return login_required

    def prepare(self):
        # shallow copy a user (as opposed to deepcopy)
        user = copy.deepcopy(current_user)
        identifier = user.get_id()
        # update current user from request
        user.remove(request.form.copy())
        logout_user()
        return (identifier, user)

    def execute(self, identifier, user):
        # insert new user
        self.auth_manager.delete(user)
        self.auth_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()



from .base import User as BaseUser
from ...base import Role as BaseRole
from ....cms import SQLContent
from ....cms import declared_attr, Column, Integer, String, Boolean, DateTime, ForeignKey, relationship

class User(BaseUser):

    @declared_attr
    def role_id(cls):
        return Column(
            Integer,
            ForeignKey('auth_role.id', ondelete='SET NULL'),
            nullable=True
        )

    @declared_attr
    def role(cls):
        return relationship('Role', foreign_keys=[cls.role_id])


class Role(BaseRole, SQLContent):
    __tablename__ = "auth_role"

    @declared_attr
    def id(cls):
        return Column(Integer, primary_key = True)

    @declared_attr
    def name(cls):
        return Column(String(50),unique=True,nullable=False)

    @declared_attr
    def privileges(cls):
        return Column(String(1500),unique=True,nullable=False)




from flask import request
from flask_login import current_user

from .base import ContentManager
from .. import exceptions, tags
from ..utils import ensure_type, ensure_callable
from ..blocks import RouteBlock


class ManageBlock(RouteBlock):

    def __init__(self, keyword, content_manager, **kwargs):
        super().__init__(keyword, **kwargs)
        ensure_type(content_manager, ContentManager, 'content_manager')
        self.content_manager = content_manager


class PrepExecBlock(ManageBlock):

    def __init__(self, keyword, content_manager, **kwargs):
        super().__init__(keyword, content_manager, **kwargs)
        ensure_callable(self.prepare, f'{self.__class__.__name__}.prepare')
        ensure_callable(self.execute, f'{self.__class__.__name__}.execute')

    def initial(self):
        return self.render()

    @property
    def default_methods(self):
        return ['GET', 'POST']

    def view(self):
        if request.method == 'POST':
            try:
                aargs = self.prepare()
            except exceptions.UserError as e:
                return self.callback(tags.USER_ERROR, e)
            except Exception as e:
                # client error
                self.client_error(e)

            try:
                return self.execute(*aargs)
            except exceptions.UserError as e:
                return self.callback(tags.USER_ERROR, e) # handle user error
            except exceptions.IntegrityError as e:
                self.content_manager.rollback() # rollback
                return self.callback(tags.INTEGRITY_ERROR, e) # handle integrity error
            except Exception as e:
                # server error: unexpected exception
                self.content_manager.rollback() # rollback
                self.server_error(e)

        try:
            return self.initial()
        except exceptions.UserError as e:
            return self.callback(tags.USER_ERROR, e)
        except Exception as e:
            # client error
            self.client_error(e)

class ContentLstBlock(ManageBlock):

    def view(self):
        c = self.content_manager.select_all()
        return self.render(data=c)


class ContentAddBlock(PrepExecBlock):

    def initial(self):
        c = self.content_manager.content_class
        return self.render(target_class=c)

    def prepare(self):
        c = self.content_manager.create(request.form.copy(), current_user)
        return (c,)

    def execute(self, target):
        # insert new user
        identifier = self.content_manager.insert(target)
        self.content_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()


class ContentModBlock(PrepExecBlock):

    def initial(self):
        c = self.content_manager.query(request.args.copy())
        return self.render(target=c)

    def prepare(self):
        c = self.content_manager.modify(request.args.copy(), request.form.copy(), current_user)
        return (c,)

    def execute(self, target):
        # insert new user
        identifier = self.content_manager.update(target)
        self.content_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()


class ContentDelBlock(PrepExecBlock):

    def initial(self):
        c = self.content_manager.query(request.args.copy())
        return self.render(target=c)

    def prepare(self):
        c = self.content_manager.query(request.args.copy())
        return (c,)

    def execute(self, target):
        # insert new user
        identifier = self.content_manager.delete(target)
        self.content_manager.commit() # commit insertion
        self.callback(tags.SUCCESS, identifier)
        return self.reroute()



class Auth:
    def auth(self, supplied_auth_data):
        if self.authd is None:
            return False
        return self.check_auth_data(supplied_auth_data)

    @classmethod
    def parse_auth_data(cls, data):
        '''
        this function should return an identifier (to create the user object) and a supplied_auth_data
        the supplied_auth_data is used in the auth(self, supplied_auth_data) method
        '''
        raise NotImplementedError(f'parse_auth_data callback on {cls.__name__} not implemented.')

    def check_auth_data(self, supplied_auth_data):
        '''
        perform authentication on user on the supplied_auth_data
        the supplied_auth_data is parsed by the parse_auth_data(cls, data) method
        '''
        raise NotImplementedError(f'check_auth_data callback on {self.__class__.__name__} not implemented.')

    def set_auth_data(self, supplied_auth_data):
        '''
        sets up the authentication data (self.authd) from the supplied auth data
        this should be called when update/create on user object (if authd is changed)
        '''
        raise NotImplementedError(f'set_auth_data callback on {self.__class__.__name__} not implemented.')

    def parse_reset_data(cls, data):
        '''
        this is used for something like password resets
        return an identifier
        '''
        raise NotImplementedError(f'parse_reset_data callback on {cls.__name__} not implemented.')

    def reset(self, data):
        '''
        this is for the user to reset
        '''
        raise NotImplementedError(f'reset callback on {self.__cls__.__name__} not implemented.')

class AuthManager:

    def select_user(self, userid):
        raise NotImplementedError(f'select_user method on {self.__class__.__name__} not implemented.')

    def parse_login(self, data):
        id, ad = self.content_class.parse_auth_data(data)
        return id, ad

    def parse_reset(self, data):
        id = self.content_class.parse_reset_data(data)
        return id

    def register(self, data):
        nu = self.content_class.register(data)
        return nu



from werkzeug.security import generate_password_hash, check_password_hash

from .. import exceptions
from ..auth import base

class Auth(base.Auth):

    def __init__(self, username, password):
        super().__init__(username)
        self.set_auth_data(password)

    @classmethod
    def parse_auth_data(cls, data):
        username = data['username']
        supplied_auth_data = data['password']
        return username, supplied_auth_data

    def check_auth_data(self, supplied_auth_data):
        return check_password_hash(self.authd, supplied_auth_data)

    def set_auth_data(self, supplied_auth_data):
        method = 'pbkdf2:sha512'
        saltlen = 16
        self.authd = generate_password_hash(supplied_auth_data, method=method, salt_length=saltlen)

    @classmethod
    def register(cls, data):
        # user register themself
        if data['password'] != data['password_confirm']:
            raise exceptions.UserError('password do not match', 400)
        nu = cls(data['username'], data['password'])
        return nu

    def renew(self, data):
        # user renew own account
        if data.get('password_new'):
            if not self.auth(data['password_old']):
                raise exceptions.UserError('invalid old password', 401)

            if data['password_new'] != data['password_confirm']:
                raise exceptions.UserError('new password do not match', 400)
            self.set_auth_data(data['password_confirm'])

    def remove(self, data):
        # user remove themselves
        if not self.auth(data['password']):
            raise exceptions.UserError('invalid password', 401)
        # do something here
        pass




from .utils import ensure_type, ensure_callable

import traceback
from jinja2.exceptions import TemplateNotFound
from flask import redirect, url_for, flash, render_template, request, abort, current_app

# late-binding vs. early binding
# https://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
def route_rewrap(wrap, routef):
    @wrap
    def f(*args):
        return routef(*args)
    return f


class RouteBlock:

    @property
    def default_url_rule(self):
        return f'/{self.keyword}'

    @property
    def default_methods(self):
        return ['GET']

    @property
    def default_access_policy(self):
        return None

    def set_custom_callback(self, tag, f):
        ensure_callable(f, 'custom_callback')
        self.callbacks[tag] = f

    # routeblock defines the default behavior
    def __init__(self, keyword, url_rule=None, template=None, reroute_to=None, reroute_external=False, reroute_kwargs={}, callbacks={}, access_policy=None, **options):
        # enforce typing
        ensure_type(keyword, str, 'keyword')
        if '.' in keyword:
            raise ValueError(f'keyword must not have dot (.) characters, not \'{keyword}\'')
        ensure_type(url_rule, str, 'url_rule', allow_none=True)
        ensure_type(template, str, 'template', allow_none=True)
        ensure_type(reroute_to, str, 'reroute', allow_none=True)
        ensure_type(reroute_external, bool, 'reroute_external')
        ensure_type(reroute_kwargs, dict, 'reroute_kwargs')
        ensure_type(callbacks, dict, 'callbacks')
        ensure_callable(access_policy, 'access_policy', allow_none=True)


        self.keyword = keyword
        self.url_rule = url_rule
        self.template = template
        self.reroute_to = reroute_to
        self.reroute_external = reroute_external
        self.reroute_kwargs = reroute_kwargs.copy()
        self.callbacks = callbacks.copy()
        self.access_policy = access_policy
        self.options = options

        if self.url_rule is None:
            self.url_rule = self.default_url_rule

        if self.template is None:
            self.template = f'{self.keyword}.html'

        if self.access_policy is None:
            self.access_policy = self.default_access_policy

        if 'methods' not in self.options:
            self.options['methods'] = self.default_methods


    def render(self, **kwargs):
        try:
            return render_template(self.template, **kwargs)
        except TemplateNotFound:
            return f'template for {self.keyword}: \'{self.template}\' not found.', 500

    def reroute(self, **kwargs):
        # reroute action
        passd = kwargs
        if isinstance(self.reroute_kwargs, dict):
            for k, v in self.reroute_kwargs.items():
                passd[k] = v

        return redirect(url_for(self.reroute_endpoint, **passd))

    def callback(self, tag, *args, **kwargs):
        if tag in self.callbacks and callable(self.callbacks[tag]):
            return self.callbacks[tag](self, *args, **kwargs)
        else:
            raise KeyError(f'custom callback for {self.keyword}.{tag} invalid')

    def finalize(self, arch_name):
        # called by the BaseArch init_app routines to ready up the block
        # ensure all configuration is done BEFORE this is called
        if not callable(self.view):
            raise NotImplementedError(f'{self.__class__.__name__} has no view method implemented.')

        if callable(self.access_policy):
            self.view = route_rewrap(self.access_policy, self.view)

        self._arch_name = arch_name

    @property
    def arch_name(self):
        return self._arch_name

    @property
    def reroute_endpoint(self):
        if self.reroute_to is None:
            raise ValueError(f'routeblock {self.__class__.__name__} has no reroute definitions')
        if '.' in self.reroute_to or self.reroute_external:
            # rerouting to external, as is
            return self.reroute_to
        else:
            # we need to append _arch_name as this is an implicit internal reroute
            return f'{self._arch_name}.{self.reroute_to}'

    def custom_reroute(self, r):
        self.reroute_external = True
        self.reroute_to = r

    def _debug(self):
        if hasattr(self, '_arch_name'):
            print(self._arch_name)
        else:
            print('unfinalized routeblock')
        print(self.keyword)
        print(self.template)
        try:
            print(self.reroute_endpoint)
        except ValueError:
            print('no reroute definition')
        print(self.reroute_kwargs)
        print(self.callbacks)
        print()

    def __str__(self):
        return self.keyword

    def __eq__(self, other):
        if isinstance(other, str):
            return self.keyword == other
        else:
            return self == other

    def client_error(self, e):
        if current_app.debug:
            traceback.print_exc()
            print('client_error', str(e))
        self.abort(400)

    def server_error(self, e):
        if current_app.debug:
            traceback.print_exc()
            raise e
        self.abort(500)

    def abort(self, code):
        abort(code)

    def flash(self, msg, cat = 'ok'):
        flash(msg, cat)

class RenderBlock(RouteBlock):

    def view(self):
        return self.render()

class RerouteBlock(RouteBlock):

    def view(self):
        return self.reroute()



# handler exceptions

from sqlalchemy.exc import IntegrityError

class UserError(Exception):
    def __init__(self, msg, code, *args):
        super().__init__(msg)
        self.msg = msg
        self.code = code
        self.args = args

INVALID_CREDS = UserError('invalid credentials', 401)




def default_success(rb, e):
    rb.flash(f'{rb.keyword} successful', 'ok')

def default_user_error(rb, e):
    rb.flash(e.msg, 'err')
    return rb.render(), e.code

def default_int_error(rb, e):
    #rb.flash('already exist', 'warn')
    rb.flash('integrity error', 'warn')
    return rb.render(), 409



def ensure_type(arg, typ, argn, allow_none = False):
    if not isinstance(arg, typ):
        if allow_none and arg is None:
            return
        raise TypeError(f'{argn} should be of instance {typ}, got {type(arg)}')

def ensure_callable(arg, argn, allow_none=False):
    if not callable(arg):
        if allow_none and arg is None:
            return
        raise TypeError(f'{argn} should be callable, got {type(arg)}')

def parse_boolean(reqform, attr):
    formval = reqform.get(attr)
    if isinstance(formval, bool):
        return formval
    elif isinstance(formval, str):
        return formval.lower() in ['on', '1', 'yes', 'true', 't']
    elif isinstance(formval, int):
        return formval > 0
    else:
        return False



from ... import base
from .... import exceptions

class ContentManager(base.ContentManager):

    def __init__(self, content_class):
        super().__init__(content_class)
        self.data = {}

    def select_all(self):
        return self.data.values()

    def select_one(self, id):
        if id in self.data:
            return self.data[id]

    def insert(self, nd):
        if nd.get_id() in self.data:
            raise exceptions.UserError(f'{self.content_class.__name__} exists.', 409)
        self.data[nd.get_id()] = nd

    def update(self, nd):
        if nd.get_id() in self.data:
            self.data[nd.get_id()] = nd
            return self.data[nd.get_id()]

    def delete(self, nd):
        if nd.get_id() in self.data:
            del self.data[nd.get_id()]

    def commit(self):
        pass

    def rollback(self):
        pass



# exports
from .base import Content as BaseContent
from .base import ContentManager as BaseContentManager

from .volatile.procmem import ContentManager as ProcMemContentManager

from .persist.sql import Content as SQLContent
from .persist.sql import ContentManager as SQLContentManager
from .persist.sql import Connection as SQLDBConnection

from sqlalchemy import Column, Integer, String, Boolean, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.declarative import declarative_base

from .blocks import ManageBlock as ContentManageBlock
from .blocks import PrepExecBlock as ContentPrepExecBlock
from .blocks import ContentLstBlock, ContentAddBlock, ContentModBlock, ContentDelBlock



from . import privileges
from .. import base, tags, callbacks
from ..utils import ensure_type
from ..cms.base import ContentManager
from ..cms.blocks import ContentLstBlock, ContentAddBlock, ContentModBlock, ContentDelBlock
from ..user.access_policies import privilege_required

class Arch(base.Arch):

    def __init__(self, user_manager, arch_name='userman', **kwargs):
        super().__init__(arch_name, **kwargs)
        ensure_type(user_manager, ContentManager, 'user_manager')

        USERLST   = 'userlst'
        USERADD   = 'useradd'
        USERMOD   = 'usermod'
        USERDEL   = 'userdel'

        rb = ContentLstBlock(USERLST, user_manager,
                access_policy=privilege_required(privileges.USERSEL))
        self.add_route_block(rb)

        rb = ContentAddBlock(USERADD, user_manager, reroute_to=USERLST,
                access_policy=privilege_required(privileges.USERADD))
        self.add_route_block(rb)

        rb = ContentModBlock(USERMOD, user_manager, reroute_to=USERLST,
                access_policy=privilege_required(privileges.USERMOD))
        self.add_route_block(rb)

        rb = ContentDelBlock(USERDEL, user_manager, reroute_to=USERLST,
                access_policy=privilege_required(privileges.USERDEL))
        self.add_route_block(rb)

        for rb in self.route_blocks.values():
            rb.set_custom_callback(tags.SUCCESS, callbacks.default_success)
            rb.set_custom_callback(tags.USER_ERROR, callbacks.default_user_error)
            rb.set_custom_callback(tags.INTEGRITY_ERROR, callbacks.default_int_error)



import json

from ..cms.base import Content

class Role(Content):
    __DELIM = ';'

    def __init__(self, name, privileges=[]):
        self.name = name
        self.privileges = '{}'
        for p in privileges:
            self.set_privilege(p, True)

    def set_privilege(self, privilege, set_1=True):
        pd = self.get_privileges()
        if set_1:
            pd[privilege] = 1
        else:
            pd.pop(privilege)
        self.privileges = json.dumps(pd)

    def has_privilege(self, privilege):
        return privilege in self.get_privileges()

    def get_privileges(self):
        return json.loads(self.privileges)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, Role):
            return all([self.get_privileges() == other.get_privileges(), self.name == other.name])
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

no_role = Role('no role', [])

class User(Content):
    '''
    ancestor of all authenticated users
    default attributes: is_authenticated, is_active, is_anonymous, userid (key), id, authd
    '''
    is_anonymous = False
    is_authenticated = False

    __contentname__ = "auth_user"

    # user identifier key. (e.g., if set to id, means user.id will identify the user)
    userid = 'id'

    def __init__(self, identifier):
        setattr(self, self.userid, identifier)
        self.is_active = True

    def get_id(self):
        if self.is_anonymous:
            return None
        else:
            return getattr(self, self.userid)

    def get_role(self):
        if hasattr(self, 'role') and self.role is not None:
            return self.role
        else:
            return no_role

    # these are different from create, update and delete, in that they are done by the users themselves
    # they are 'self' methods, as opposed to someone else creating/updating and deleting
    @classmethod
    def register(cls, data):
        raise NotImplementedError(f'register method on {cls.__name__} not implemented')

    def renew(self, data):
        raise NotImplementedError(f'renew method on {self.__class__.__name__} not implemented')

    def remove(self, data):
        raise NotImplementedError(f'remove method on {self.__class__.__name__} not implemented')



from .base import User
from .role import User as UserWithRole
from .role import Role

from ....auth.base import Auth, AuthManager
from ....cms.persist.sql import ContentManager

class UserManager(AuthManager, ContentManager):

    def __init__(self, auth_class, db_conn, user_class=User):
        if not issubclass(auth_class, Auth):
            raise TypeError(f'{auth_class} should be a subclass of {Auth}.')
        if not issubclass(user_class, User):
            raise TypeError(f'{user_class} should be a subclass of {User}.')

        class AuthUser(auth_class, user_class, db_conn.orm_base):
            pass

        super().__init__(AuthUser, db_conn)

    def select_user(self, userid):
        return self.content_class.query.filter(
            getattr(self.content_class, self.content_class.userid) == userid
        ).first()



from setuptools import find_packages, setup

_version = '0.0.6'

setup(
    name='flask-arch',
    version=_version,
    description='a modular architecture project for bootstrapping on the flask web development microframework',
    packages=find_packages(),
    author='Chia Jason',
    author_email='chia_jason96@live.com',
    url='https://github.com/toranova/flask-arch/',
    download_url='https://github.com/ToraNova/flask-arch/archive/refs/tags/v%s.tar.gz' % _version,
    license='MIT',
    include_package_data=True,
    zip_safe=False,
    keywords = ['Flask'],
    install_requires=[
        'flask',
        'flask-login',
        'sqlalchemy',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



# this the shared class for most of the arch on flask-arch
# ported from vials project to flask-arch, 2022 feb 21
# author: toranova
# mailto: chia_jason96@live.com

from .base import Arch as BaseArch
from .blocks import RouteBlock



#--------------------------------------------------
# fsqlite.py (sqlalchemy ORM base)
# this file is static and should not be tampered with
# it initializes the required base models for the db engine
# introduced 8/12/2018
# migrated from rapidflask to miniflask (22 Jul 2020)
# migrated from miniflask to vials project (29 Nov 2020)
# migrated from vials project to the flask-arch project (21 Feb 2022)
# ToraNova 2022
# chia_jason96@live.com
#--------------------------------------------------

import json
import datetime
from sqlalchemy import create_engine, MetaData, inspect
from sqlalchemy.orm import scoped_session, sessionmaker, relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy import Column, Integer, String, Boolean, DateTime, ForeignKey
from flask_login import current_user

from ... import base

def make_session(engine, base):
    '''create a session and bind the Base query property to it'''
    sess =  scoped_session(sessionmaker(autocommit=False,autoflush=False,bind=engine))
    base.query = sess.query_property()
    return sess

class Connection:

    def __init__(self, db_uri, orm_base):
        self.uri = db_uri
        self.engine = create_engine(self.uri)
        self.orm_base = orm_base
        self.session = make_session(self.engine, self.orm_base)

    def configure_teardown(self, app):

        @app.teardown_appcontext
        def teardown(exception=None):
            self.session.remove()

class Content(base.Content):

    def as_json(self):
        return json.dumps(self.as_dict())

    def as_dict(self):
        # dump all table into a dictionary
        od = {c.name: (getattr(self, c.name)) for c in self.__table__.columns}
        for k, v in od.items():
            # convert dates to isoformat
            if isinstance(v, datetime.datetime):
                od[k] = v.isoformat()
        return od

    @declared_attr
    def created_on(cls):
        return Column(DateTime()) # date of content creation

    @declared_attr
    def creator_id(cls):
        return Column(
            Integer,
            ForeignKey('auth_user.id', ondelete='SET NULL'),
            nullable=True
        )

    @declared_attr
    def created_by(cls):
        return relationship('AuthUser', foreign_keys=[cls.creator_id])

    @declared_attr
    def modified_on(cls):
        return Column(DateTime()) # date of content update

    @declared_attr
    def modifier_id(cls):
        return Column(
            Integer,
            ForeignKey('auth_user.id', ondelete='SET NULL'),
            nullable=True
        )

    @declared_attr
    def modified_by(cls):
        return relationship('AuthUser', foreign_keys=[cls.modifier_id])


class ContentManager(base.ContentManager):

    @property
    def session(self):
        return self.db_conn.session

    @property
    def database_uri(self):
        return self.db_conn.uri

    def __init__(self, content_class, db_conn):
        super().__init__(content_class)
        if not issubclass(content_class, Content):
            raise TypeError(f'{content_class} should be a subclass of {Content}.')

        if not isinstance(db_conn, Connection):
            raise TypeError(f'{db_conn} should be a subclass of {Connection}.')

        self.db_conn = db_conn
        self.tablename = self.content_class.__tablename__

    # create table if not exist on dburi
    def create_table(self):
        engine = create_engine(self.database_uri)
        self.content_class.__table__.create(engine, checkfirst=True)
        engine.dispose() #house keeping

    # check if table exists in dburi
    def table_exists(self):
        engine = create_engine(self.database_uri)
        ins = inspect(engine)
        res = self.content_class.__tablename__ in ins.get_table_names()
        engine.dispose()
        return res

    def select_all(self):
        return self.content_class.query.all()

    def select_one(self, id):
        return self.content_class.query.filter(self.content_class.id == id).first()

    # insert/update/delete queries
    def insert(self, nd):
        # insert a new content
        self.session.add(nd)

    def update(self, nd):
        # update a content
        self.session.add(nd)

    def delete(self, nd):
        # delete a content
        self.session.delete(nd)

    # persistence method
    def commit(self):
        # persist changes and synchronize
        self.session.commit()

    def rollback(self):
        # rollback changes (encountered an exception)
        self.session.rollback()



# basic authentication (username, password)
# no database systems, users defined by python scripts

from flask import abort
from flask_login import LoginManager, login_required

from .. import base, tags, callbacks
from ..auth.base import AuthManager
from ..auth.blocks import LoginBlock, LogoutBlock, RegisterBlock, RenewBlock, ResetBlock, RemoveBlock
from ..utils import ensure_type
from ..blocks import RenderBlock

# basic.Arch
class Arch(base.Arch):

    def __init__(self, user_manager, arch_name='auth', **kwargs):
        '''
        initialize the architecture for the flask_arch
        templ is a dictionary that returns user specified templates to user on given routes
        reroutes is a dictionary that reroutes the user after certain actions on given routes
        '''
        super().__init__(arch_name, **kwargs)
        ensure_type(user_manager, AuthManager, 'user_manager')

        LOGIN   = 'login'
        LOGOUT  = 'logout'
        PROFILE = 'profile'
        INSERT  = 'register'
        UPDATE  = 'renew'
        RESET  = 'reset'
        DELETE  = 'remove'

        rb = RenderBlock(PROFILE, access_policy=login_required)
        self.add_route_block(rb)

        rb = LoginBlock(LOGIN, user_manager, reroute_to=PROFILE)
        self.add_route_block(rb)

        rb = LogoutBlock(LOGOUT, user_manager, reroute_to=LOGIN)
        self.add_route_block(rb)

        rb = RegisterBlock(INSERT, user_manager, reroute_to=LOGIN)
        self.add_route_block(rb)

        rb = RenewBlock(UPDATE, user_manager, reroute_to=PROFILE)
        self.add_route_block(rb)

        rb = ResetBlock(RESET, user_manager, reroute_to=LOGIN)
        self.add_route_block(rb)

        rb = RemoveBlock(DELETE, user_manager, reroute_to=LOGIN)
        self.add_route_block(rb)

        for rb in self.route_blocks.values():
            rb.set_custom_callback(tags.SUCCESS, callbacks.default_success)
            rb.set_custom_callback(tags.USER_ERROR, callbacks.default_user_error)
            rb.set_custom_callback(tags.INTEGRITY_ERROR, callbacks.default_int_error)

        self.login_manager = LoginManager()

        @self.login_manager.unauthorized_handler
        def unauthorized():
            abort(401)

        @self.login_manager.user_loader
        def loader(userid):
            user = user_manager.select_user(userid)
            if user is None:
                return None
            user.is_authenticated = True
            return user

    def init_app(self, app):
        super().init_app(app)

        self.login_manager.init_app(app)



# this the shared class for most of the arch on flask-arch
# ported from vials project to flask-arch, 2022 feb 21
# author: toranova
# mailto: chia_jason96@live.com

from flask import Blueprint
from .blocks import RouteBlock
from .utils import ensure_type

def valid_override(keyword, cdict, strict_type):
    if keyword in cdict:
        ensure_type(cdict[keyword], strict_type, f'[{keyword}] = {cdict[keyword]}')
        return True
    return False

class Arch:

    def configure(self, keyword):
        if keyword not in self.route_blocks:
            raise KeyError('routeblock \'{keyword}\' cannot be configured because it does not exist.')
        return self.route_blocks[keyword]

    def add_route_blocks(self, rblist):
        for rb in rblist:
            self.add_route_block(rb)

    def add_route_block(self, rb, override=False):
        ensure_type(rb, RouteBlock, 'routeblock')
        if rb.keyword in self.route_blocks and not override:
            raise KeyError('routeblock \'{rb.keyword}\' already exists! set override=True to override.')
        self.route_blocks[rb.keyword] = rb

    def _debug(self):
        for rb in self.route_blocks.values():
            print(self.url_prefix)
            rb._debug()

    def init_app(self, app):

        for rb in self.route_blocks.values():
            if not isinstance(rb, RouteBlock):
                continue

            if rb.keyword in self.routes_disabled:
                continue

            # user overrides template
            if valid_override(rb.keyword, self.custom_templates, str):
                rb.template = self.custom_templates[rb.keyword]

            # user overrides reroute
            if valid_override(rb.keyword, self.custom_reroutes, str):
                rb.custom_reroute(self.custom_reroutes[rb.keyword])

            # user overrides reroute_kwargs
            if valid_override(rb.keyword, self.custom_reroutes_kwargs, dict):
                for argk, argv in self.custom_reroutes_kwargs[rb.keyword].items():
                    rb.reroute_kwargs[argk] = argv

            # user override custom_callbacks
            if valid_override(rb.keyword, self.custom_callbacks, dict):
                for tag, user_cb in self.custom_callbacks[rb.keyword].items():
                    if not callable(user_cb):
                        raise TypeError(f'custom_callbacks[\'{rb.keyword}\'][\'{tag}\'] not callable')
                    rb.callbacks[tag] = user_cb

            rb.finalize(self.name)
            self.bp.add_url_rule(rb.url_rule, rb.keyword, rb.view, **rb.options)

        app.register_blueprint(self.bp)
        #self._debug()  # enable for debugging

    def __init__(self, arch_name, custom_templates = {}, custom_reroutes = {}, custom_reroutes_kwargs = {}, custom_callbacks = {}, custom_url_prefix = None, routes_disabled=[]):
        '''
        arch_name - name of the architecture
        route_blocks - the route blocks to initialize and configure
        custom_templates, custom_reroutes, custom_reroutes_kwargs, custom_callbacks - user overrides
        url_prefix - url prefix of a blueprint generated. use / to have NO prefix, leave it at None to default to /<arch_name>
        '''

        ensure_type(arch_name, str, 'arch_name')
        ensure_type(custom_templates, dict, 'custom_templates')
        ensure_type(custom_reroutes, dict, 'custom_reroutes')
        ensure_type(custom_reroutes_kwargs, dict, 'custom_reroutes_kwargs')
        ensure_type(custom_callbacks, dict, 'custom_callbacks')
        ensure_type(custom_url_prefix, str, 'custom_url_prefix', allow_none=True)
        ensure_type(routes_disabled, list, 'routes_disabled')

        self.name = arch_name
        self.custom_templates = custom_templates.copy()
        self.custom_reroutes = custom_reroutes.copy()
        self.custom_reroutes_kwargs = custom_reroutes_kwargs.copy()
        self.custom_callbacks = custom_callbacks.copy()
        self.custom_url_prefix = custom_url_prefix
        self.routes_disabled = routes_disabled.copy()

        self.route_blocks = {}

        if self.custom_url_prefix is None:
            self.url_prefix = '/%s' % self.name
        elif self.custom_url_prefix == '/':
            self.url_prefix = None
        else:
            self.url_prefix = self.custom_url_prefix

        self.bp = Blueprint(self.name, __name__, url_prefix = self.url_prefix)
        self.blueprint = self.bp # alias
